@@ -1,8 +1,10 @@
+#!/usr/bin/env python3
 # -*- coding: utf-8 -*-
-"""
-Broker Utilities SuperClass
-
-"""
+'''
+Module: brokers.broker_utils
+Purpose: Broker utilities superclass
+Author: Kieran Mackle
+'''
 
 import pandas as pd
 import numpy as np
@@ -75,7 +77,7 @@ class BrokerUtils:
             else:
                 quote_risk      = amount_risked / HCF
                 price_per_pip   = quote_risk / pip_stop_distance
-                units           = round(price_per_pip / pip_value)
+                units           = price_per_pip / pip_value
         
         return units
     
