@@ -1,10 +1,10 @@
 #!/usr/bin/env python3
 # -*- coding: utf-8 -*-
-"""
-Created on Wed Jun 23 10:19:44 2021
-
-@author: kieran
-"""
+'''
+Module: lib.email_manager
+Purpose: Email utility module
+Author: Kieran Mackle
+'''
 
 from getopt import getopt
 import sys
@@ -18,7 +18,7 @@ def main(uo_dict):
     file_dir    = os.path.join(email_dir, '../logfiles')
     file_path   = os.path.join(file_dir, uo_dict["filename"])
     
-    global_config   = read_yaml(email_dir + '../config/GLOBAL.yaml')
+    global_config   = read_yaml(email_dir + '/../config/GLOBAL.yaml')
     host_email      = global_config["EMAILING"]["HOST_ACCOUNT"]
     mailing_list    = global_config["EMAILING"]["MAILING_LIST"]
     
@@ -49,4 +49,4 @@ if __name__ == '__main__':
         
     uo_dict = {'filename': filename}
 
-    main(uo_dict)
\ No newline at end of file
+    main(uo_dict)
