@@ -10,7 +10,7 @@ from finta import TA
 import numpy as np
 import pandas as pd
 
-
+''' -------------------------- PRICE INDICATORS --------------------------- '''
 def supertrend(data, period = 10, ATR_multiplier = 3.0, source=None):
     ''' Based on the SuperTrend indicator by KivancOzbilgic on TradingView '''
     
@@ -150,90 +150,6 @@ def atr(data, period=14):
     return atr
 
 
-def rsi():
-    return
-
-def crossover(list_1, list_2):
-    ''' 
-    Returns a list of length len(list_1) with 1 when list_1 crosses above
-    list_2 and -1 when list_1 crosses below list_2.
-    '''
-    
-    sign_list = []
-    for i in range(len(list_1)):
-        if np.isnan(list_1[i]):
-            sign_list.append(np.nan)
-        else:
-            difference = list_1[i] - list_2[i]
-            if difference < 0:
-                sign_list.append(-1)
-            else:
-                sign_list.append(1)
-    
-    crossover_list = [0]
-    
-    for i in range(1, len(sign_list)):
-        if sign_list[i] - sign_list[i-1] != 0:
-            val = sign_list[i]
-        else:
-            val = 0
-        
-        crossover_list.append(val)
-
-    return crossover_list
-
-
-def cross_values(a, b, ab_crossover):
-    cross_point_list = [0]
-    last_cross_point = 0
-    for i in range(1, len(ab_crossover)):
-        if ab_crossover[i] != 0:
-            i0 = 0
-            m_a = a[i] - a[i-1]
-            m_b = b[i] - b[i-1]
-            ix = (b[i-1] - a[i-1])/(m_a-m_b) + i0
-            
-            cross_point = m_a*(ix - i0) + a[i-1]
-            
-            last_cross_point = cross_point
-            
-        else:
-            cross_point = last_cross_point #0
-        
-        cross_point_list.append(cross_point)
-    
-    # Replace nans with 0
-    cross_point_list = [0 if x!=x  else x for x in cross_point_list]
-    
-    return cross_point_list
-
-
-def candles_between_crosses(cross_list):
-    '''
-    Returns candles since last cross
-    
-    
-    Behaviour:
-    in:  [0, 0, 1, 0, 0, 0, -1, 0, 0, 0, 0, 0, 1]
-    out: [1, 2, 0, 1, 2, 3,  0, 1, 2, 3, 4, 5, 0]
-    '''
-    
-    count = 0
-    count_list = []
-    
-    for i in range(len(cross_list)):
-
-        if cross_list[i] == 0:
-            # Change in signal - reset count
-            count += 1
-        else:
-            count = 0
-        
-        count_list.append(count)
-    
-    return count_list
-
-
 def bullish_engulfing(data, detection = None):
     ''' Bullish engulfing pattern detection. '''
     
@@ -405,82 +321,303 @@ def heikin_ashi(data):
     return ha_data
  
 
-def find_swings(data, use_body = False):
-    '''
-        Locates the recent swings in price and returns a rolling list of 
-        swing prices.
-    '''
-    
-    # TODO - include resampling option to find swings on higher timeframe
-    
-    # use_body flag currently not implemented. Idea is to use body close/open
-    # instead of high/low (wicks) of candle for swings.
-    
-    # oc2     = (data.Open.values + data.Close.values)/2
-    hl2     = (data.High.values + data.Low.values)/2
-    n       = 2
-    EMA     = ema(hl2, n)
-    
-    grad    = np.gradient(EMA)
-    
-    swings   = np.zeros(len(grad))
-    for i in range(1, len(grad)):
-        if not np.isnan(grad[i-1]):
-            if np.sign(grad[i]) != np.sign(grad[i-1]):
-                swings[i] = -np.sign(grad[i])
 
-    swing_df = pd.DataFrame(data=swings, index=data.index, columns=['swing'])
+
+''' ------------------------ UTILITY INDICATORS --------------------------- '''
+
+def crossover(list_1, list_2):
+    ''' 
+    Returns a list of length len(list_1) with 1 when list_1 crosses above
+    list_2 and -1 when list_1 crosses below list_2.
+    '''
     
+    sign_list = []
+    for i in range(len(list_1)):
+        if np.isnan(list_1[i]):
+            sign_list.append(np.nan)
+        else:
+            difference = list_1[i] - list_2[i]
+            if difference < 0:
+                sign_list.append(-1)
+            else:
+                sign_list.append(1)
+    
+    crossover_list = [0]
+    
+    for i in range(1, len(sign_list)):
+        if sign_list[i] - sign_list[i-1] != 0:
+            val = sign_list[i]
+        else:
+            val = 0
+        
+        crossover_list.append(val)
+
+    return crossover_list
+
+
+def cross_values(a, b, ab_crossover):
+    cross_point_list = [0]
+    last_cross_point = 0
+    for i in range(1, len(ab_crossover)):
+        if ab_crossover[i] != 0:
+            i0 = 0
+            m_a = a[i] - a[i-1]
+            m_b = b[i] - b[i-1]
+            ix = (b[i-1] - a[i-1])/(m_a-m_b) + i0
+            
+            cross_point = m_a*(ix - i0) + a[i-1]
+            
+            last_cross_point = cross_point
+            
+        else:
+            cross_point = last_cross_point #0
+        
+        cross_point_list.append(cross_point)
+    
+    # Replace nans with 0
+    cross_point_list = [0 if x!=x  else x for x in cross_point_list]
+    
+    return cross_point_list
+
+
+def candles_between_crosses(cross_list):
+    '''
+    Returns candles since last cross
+    
+    
+    Behaviour:
+    in:  [0, 0, 1, 0, 0, 0, -1, 0, 0, 0, 0, 0, 1]
+    out: [1, 2, 0, 1, 2, 3,  0, 1, 2, 3, 4, 5, 0]
+    '''
+    
+    count = 0
+    count_list = []
+    
+    for i in range(len(cross_list)):
+
+        if cross_list[i] == 0:
+            # Change in signal - reset count
+            count += 1
+        else:
+            count = 0
+        
+        count_list.append(count)
+    
+    return count_list
+
+def find_swings(data, data_type='ohlc', n = 2):
+    '''
+    Locates swings in the inputted data and returns a dataframe.
+    
+    Parameters:
+        data: an OHLC dataframe of price, or an array/list of data from an 
+        indicator.
+        
+        data_type: specify 'ohlc' when data is OHLC, or 'other' when inputting
+        an indicator.
+    '''
+    
+    # Prepare data 
+    if data_type == 'ohlc':
+        # Find swings in OHLC price data
+        # hl2     = (data.Open.values + data.Close.values)/2
+        hl2     = (data.High.values + data.Low.values)/2
+        swing_data = ema(hl2, n)
+        
+        low_data = data.Low.values
+        high_data = data.High.values
+        
+    else:
+        # Find swings in alternative data source
+        swing_data = data
+        low_data = data
+        high_data = data
+    
+    # Calculate slope of data and points where slope changes
+    grad = [swing_data[i] - swing_data[i-1] for i in range(len(swing_data))]
+    swings = np.where(np.sign(grad) != np.sign(np.roll(grad,1)), -np.sign(grad), 0)
+
+    # Construct columns
     low_list    = []
     high_list   = []
     for i in range(len(data)):
-        if swing_df.swing[i] == -1:
-            low_list.append(min(data.Low.values[i-1:i+2]))
+        if swings[i] == -1:
+            # Down swing - find min price in the vicinity 
             high_list.append(0)
-        elif swing_df.swing[i] == 1:
-            high_list.append(max(data.High.values[i-1:i+2]))
+            low_list.append(min(low_data[i-2:i]))
+            
+        elif swings[i] == 1:
+            # Up swing - find max price in the vicinity
+            high_list.append(max(high_data[i-2:i]))
             low_list.append(0)
+            
         else:
+            # Price movement
             low_list.append(0)
             high_list.append(0)
     
-    swings_list     = merge_signals(low_list, high_list)
-    last_swing      = rolling_signal_list(swings_list)
-    last_swing[0:n] = list(data.High.values[0:n])
+    trend = rolling_signal_list(-swings)
+    swings_list = merge_signals(low_list, high_list)
+    last_swing = rolling_signal_list(swings_list)
+    last_swing[0:n] = list(high_data[0:n])
     
     # Need to return both a last swing low and last swing high list
-    last_low        = rolling_signal_list(low_list)
-    last_low[0:n]   = list(data.Low.values[0:n])
-    last_high       = rolling_signal_list(high_list)
-    last_high[0:n]  = list(data.High.values[0:n])
+    last_low = rolling_signal_list(low_list)
+    last_low[0:n] = list(low_data[0:n])     # Fill start of data
+    last_high = rolling_signal_list(high_list)
+    last_high[0:n] = list(high_data[0:n])   # Fill start of data
     
-    swings          = pd.DataFrame(data={'Highs': last_high, 
-                                         'Lows' : last_low,
-                                         'Last' : last_swing},
-                                   index = data.index)
+    swing_df = pd.DataFrame(data={'Highs': last_high, 
+                                  'Lows' : last_low,
+                                  'Last' : last_swing,
+                                  'Trend': trend},
+                            index = data.index)
     
-    return swings
+    return swing_df
     
+def classify_swings(swing_df, tol=0):
+    ''' 
+    Classify a dataframe of swings (from find_swings) into higher-high, 
+    lower-high, higher-low and lower-low.
+    
+    Parameters:
+        swing_df: the dataframe outputted from find_swings.
+        
+        tol: parameter to control strength of levels detected.
+    '''
+    
+    # Create copy of swing dataframe
+    swing_df = swing_df.copy()
+    
+    new_level = np.where(swing_df.Last != swing_df.Last.shift(), 1, 0)
+
+    candles_since_last = candles_between_crosses(new_level)
+    
+    # Add column 'candles since last swing' CSLS
+    swing_df['CSLS'] = candles_since_last
+    
+    # Find strong Support and Resistance zones
+    swing_df['Support'] = (swing_df.CSLS > tol) & (swing_df.Trend == 1)
+    swing_df['Resistance'] = (swing_df.CSLS > tol) & (swing_df.Trend == -1)
+    
+    # Find higher highs and lower lows
+    swing_df['Strong_lows'] = swing_df['Support'] * swing_df['Lows'] # Returns high values when there is a strong support
+    swing_df['Strong_highs'] = swing_df['Resistance'] * swing_df['Highs'] # Returns high values when there is a strong support
+    
+    # Remove duplicates to preserve indexes of new levels
+    swing_df['FSL'] = unroll_signal_list(swing_df['Strong_lows']) # First of new strong lows
+    swing_df['FSH'] = unroll_signal_list(swing_df['Strong_highs']) # First of new strong highs
+    
+    # Now compare each non-zero value to the previous non-zero value.
+    low_change = np.sign(swing_df.FSL) * (swing_df.FSL - swing_df.Strong_lows.replace(to_replace=0, method='ffill').shift())
+    high_change = np.sign(swing_df.FSH) * (swing_df.FSH - swing_df.Strong_highs.replace(to_replace=0, method='ffill').shift())
+    
+    swing_df['LL'] = np.where(low_change < 0, True, False)
+    swing_df['HL'] = np.where(low_change > 0, True, False)
+    swing_df['HH'] = np.where(high_change > 0, True, False)
+    swing_df['LH'] = np.where(high_change < 0, True, False)
+    
+    return swing_df
+
+def detect_divergence(classified_price_swings, classified_indicator_swings, tol=2):
+    '''
+    Detects divergence between price swings and swings in an indicator.
+    
+    Parameters:
+        classified_price_swings: output from classify_swings using OHLC data.
+        
+        classified_indicator_swings: output from classify_swings using indicator data.
+        
+        tol: number of candles which conditions must be met within. 
+    '''
+    
+    regular_bullish = []
+    regular_bearish = []
+    hidden_bullish = []
+    hidden_bearish = []
+    
+    for i in range(len(classified_price_swings)):
+        # Look backwards in each
+        
+        # REGULAR BULLISH DIVERGENCE
+        if sum(classified_price_swings['LL'][i-tol:i]) + sum(classified_indicator_swings['HL'][i-tol:i]) > 1:
+            regular_bullish.append(True)
+        else:
+            regular_bullish.append(False)
+        
+        # REGULAR BEARISH DIVERGENCE
+        if sum(classified_price_swings['HH'][i-tol:i]) + sum(classified_indicator_swings['LH'][i-tol:i]) > 1:
+            regular_bearish.append(True)
+        else:
+            regular_bearish.append(False)
+        
+        # HIDDEN BULLISH DIVERGENCE
+        if sum(classified_price_swings['HL'][i-tol:i]) + sum(classified_indicator_swings['LL'][i-tol:i]) > 1:
+            hidden_bullish.append(True)
+        else:
+            hidden_bullish.append(False)
+        
+        # HIDDEN BEARISH DIVERGENCE
+        if sum(classified_price_swings['LH'][i-tol:i]) + sum(classified_indicator_swings['HH'][i-tol:i]) > 1:
+            hidden_bearish.append(True)
+        else:
+            hidden_bearish.append(False)
+        
+        divergence = pd.DataFrame(data = {'regularBull': unroll_signal_list(regular_bullish),
+                                          'regularBear': unroll_signal_list(regular_bearish),
+                                          'hiddenBull': unroll_signal_list(hidden_bullish),
+                                          'hiddenBear': unroll_signal_list(hidden_bearish)})
+        
+    return divergence
+
+def autodetect_divergence(ohlc, indicator_data):
+    '''
+    Wrapper method to automatically detect divergence from inputted OHLC price 
+    data and indicator data.
+    
+    This method calls:
+        find_swings()
+        classify_swings()
+        detect_divergence()
+    
+    Parameters:
+        ohlc: dataframe of OHLC data
+        
+        indicator data: array of indicator data
+    '''
+    
+    # Price swings
+    price_swings = find_swings(ohlc)
+    price_swings_classified = classify_swings(price_swings)
+    
+    # Indicator swings 
+    indicator_swings = find_swings(indicator_data, data_type='other')
+    indicator_classified = classify_swings(indicator_swings)
+    
+    # Detect divergence
+    divergence = detect_divergence(price_swings_classified, indicator_classified)
+    
+    return divergence
 
 def rolling_signal_list(signals):
-        ''' 
-            Returns a list which maintains the previous signal, until a new 
-            signal is given.
-            
-            [0,1,0,0,0,-1,0,0,1,0,0] ->  [0,1,1,1,1,-1,-1,-1,1,1,1]
-            
-        '''
+    ''' 
+        Returns a list which maintains the previous signal, until a new 
+        signal is given.
         
-        rolling_signals = [0]
-        last_signal     = rolling_signals[0]
+        [0,1,0,0,0,-1,0,0,1,0,0] ->  [0,1,1,1,1,-1,-1,-1,1,1,1]
         
-        for i in range(1, len(signals)):
-            if signals[i] != 0:
-                last_signal = signals[i]
-            
-            rolling_signals.append(last_signal)
+    '''
+    
+    rolling_signals = [0]
+    last_signal     = rolling_signals[0]
+    
+    for i in range(1, len(signals)):
+        if signals[i] != 0:
+            last_signal = signals[i]
         
-        return rolling_signals
+        rolling_signals.append(last_signal)
+    
+    return rolling_signals
 
 def unroll_signal_list(signals):
     ''' Unrolls a signal list. '''
@@ -498,7 +635,7 @@ def merge_signals(signal_1, signal_2):
          Returns a single signal list which has merged two signal lists. 
      '''
      
-     merged_signal_list = signal_1
+     merged_signal_list = signal_1.copy()
      
      for i in range(len(signal_1)):
          if signal_2[i] != 0:
@@ -715,36 +852,3 @@ def last_level_touched(data, grid):
     
     return levels_touched
 
-
-# def TDI(data):
-#     rsiPeriod = input(11, minval = 1, title = "RSI Period")
-#     bandLength = input(31, minval = 1, title = "Band Length")
-#     lengthrsipl = input(1, minval = 0, title = "Fast MA on RSI")
-#     lengthtradesl = input(9, minval = 1, title = "Slow MA on RSI")
-    
-#     src = close                                                             // Source of Calculations (Close of Bar)
-#     r = rsi(src, rsiPeriod)                                                 // RSI of Close
-#     ma = sma(r, bandLength)                                                 // Moving Average of RSI [current]
-#     offs = (1.6185 * stdev(r, bandLength))                                  // Offset
-#     up = ma + offs                                                          // Upper Bands
-#     dn = ma - offs                                                          // Lower Bands
-#     mid = (up + dn) / 2                                                     // Average of Upper and Lower Bands
-#     fastMA = sma(r, lengthrsipl)                                            // Moving Average of RSI 2 bars back
-#     slowMA = sma(r, lengthtradesl)                                          // Moving Average of RSI 7 bars back
-    
-#     hline(30)                                                               // Oversold
-#     hline(50)                                                               // Midline
-#     hline(70)                                                               // Overbought
-    
-#     upl = plot(up, "Upper Band", color = blue)                              // Upper Band
-#     dnl = plot(dn, "Lower Band", color = blue)                              // Lower Band
-#     midl = plot(mid, "Middle of Bands", color = orange, linewidth = 2)      // Middle of Bands
-    
-#     plot(slowMA, "Slow MA", color=green, linewidth=2)                       // Plot Slow MA
-#     plot(fastMA, "Fast MA", color=red, linewidth=2)                         // Plot Fast MA
-    
-#     fill(upl, midl, red, transp=90)                                         // Fill Upper Half Red
-#     fill(midl, dnl, green, transp=90)                                       // Fill Lower Half Green
-
-
-
